import javax.swing.SwingUtilities;
import javax.swing.JFrame;
import java.io.*;
import java.net.*;
import java.util.*;

public class Satellite extends Thread
{
  protected InetAddress planetIP = null;
  protected int planetPort = 0;
  protected InetAddress sunIP;
  protected int sunPort;
  protected byte[] emailAndId;
  protected boolean isTerminated = false;
  protected int sunIntervalSec = 59;
  protected int planetIntervalSec = 1;
  protected byte[] sunPassword;
  protected byte[] planetPassword;
  protected Random rand;
  
  protected static int PROBE_HISTORY_LENGTH = 100;
  protected static int PROBE_HISTORY_MAX_INCLUDE = 20;
  protected static int SUN_PASSWORD_VALID_MILLIS = 120000;
  
  public Satellite(String email, String identifier, InetAddress sunIP, int sunPort) throws UnsupportedEncodingException {
    this.sunIP = sunIP;
    this.sunPort = sunPort;
    this.emailAndId = new byte[email.length() + identifier.length() + 2];
    this.planetPassword = new byte[0];
    this.rand = new Random();
    this.sunPassword = new byte[8];
    rand.nextBytes(sunPassword);
    System.arraycopy(email.getBytes("UTF-8"), 0, emailAndId, 0, email.length());
    System.arraycopy(identifier.getBytes("UTF-8"), 0, emailAndId, email.length()+1, identifier.length());
  }

  protected static int decodeShort(byte[] b, int offset) {
    return ((b[offset]&0xFF)<<8) | (b[offset+1]&0xFF);
  }
  
  protected static void encodeShort(int value, byte[] b, int offset) {
    b[offset] = (byte)((value>>8)&0xFF);
    b[offset+1] = (byte)(value&0xFF);
  }
  
  protected static void encodeInt(int value, byte[] b, int offset) {
    b[offset] = (byte)((value>>24)&0xFF);
    b[offset+1] = (byte)((value>>16)&0xFF);
    b[offset+2] = (byte)((value>>8)&0xFF);
    b[offset+3] = (byte)(value&0xFF);
  }
  
  public void terminate() {
    isTerminated = true;
    interrupt();
  }
  
  public void start() {
    try {
      DatagramSocket ds = new DatagramSocket(0); /* any port is fine */
      DatagramPacket pack = new DatagramPacket(new byte[65536], 65536);
      long nextPlanetReportMillis = System.currentTimeMillis();
      long nextSunReportMillis = System.currentTimeMillis();
      long sunPasswordChosen = System.currentTimeMillis();
      byte[] outputBuffer = new byte[65536];
      int[] probeHistory = new int[PROBE_HISTORY_LENGTH];
      int probeIdsInHistory = 0;
      int totalBytesSent = 0;
      int totalBytesReceived = 0;
      
      while (!isTerminated) {
        try {
          long now = System.currentTimeMillis();
          if (now >= nextPlanetReportMillis) {
            nextPlanetReportMillis = now + planetIntervalSec * 1000;
            if ((planetIP != null) && (planetPort != 0)) {
              byte[] outbuf = new byte[8+planetPassword.length];
              System.arraycopy(new byte[] { -1, -2, 0x53, 0x41, 0x54, 0x4C, 0x03, 0x01 }, 0, outbuf, 0, 8);
              System.arraycopy(planetPassword, 0, outbuf, 8, planetPassword.length);
              totalBytesSent += 28+outbuf.length;
              ds.send(new DatagramPacket(outbuf, outbuf.length, planetIP, planetPort));
            }
          }
          
          if (now >= nextSunReportMillis) {
            nextSunReportMillis = now + sunIntervalSec * 1000;
            if (now > (sunPasswordChosen+SUN_PASSWORD_VALID_MILLIS)) {
              rand.nextBytes(sunPassword);
              sunPasswordChosen = now;
            }

            byte[] outbuf = new byte[32+emailAndId.length+planetPassword.length];
            System.arraycopy(new byte[] { 0x53, 0x41, 0x54, 0x4C, 0x03, 0x01 }, 0, outbuf, 0, 6);
            System.arraycopy(sunPassword, 0, outbuf, 6, 8);
            encodeShort(sunIntervalSec, outbuf, 14);
            encodeShort(planetIntervalSec, outbuf, 16);
            if (planetIP != null)
              System.arraycopy(planetIP.getAddress(), 0, outbuf, 18, 4);
            encodeShort(planetPort, outbuf, 22);
            encodeInt(totalBytesSent, outbuf, 24);
            encodeInt(totalBytesReceived, outbuf, 28);
            System.arraycopy(emailAndId, 0, outbuf, 32, emailAndId.length);
            System.arraycopy(planetPassword, 0, outbuf, 32+emailAndId.length, planetPassword.length);
            totalBytesSent += 28+outbuf.length;
            ds.send(new DatagramPacket(outbuf, outbuf.length, sunIP, sunPort));
          }
          
          ds.setSoTimeout(Math.max(1, Math.min((int)(nextPlanetReportMillis - now), (int)(nextSunReportMillis - now))));
          
          try {
            ds.receive(pack);
            long recvTS = System.nanoTime()/1000;
            
            if ((pack.getPort() == sunPort) && pack.getAddress().equals(sunIP) && (pack.getLength() >= 24)) {
              boolean passwordOK = true;
              for (int i=0; i<8; i++) {
                if (pack.getData()[6+i] != sunPassword[i])
                  passwordOK = false;
              }
                  
              if (passwordOK) {
                totalBytesReceived += 28 + pack.getLength();
                sunIntervalSec = Math.max(1, decodeShort(pack.getData(), 14));
                planetIntervalSec = Math.max(1, decodeShort(pack.getData(), 16));
                planetPort = decodeShort(pack.getData(), 22);
                planetPassword = new byte[pack.getLength()-24];
                System.arraycopy(pack.getData(), 24, planetPassword, 0, planetPassword.length);
                
                byte[] plip = new byte[4];
                System.arraycopy(pack.getData(), 18, plip, 0, 4);
                if ((plip[0] == 0) && (plip[1] == 0) && (plip[2] == 0) && (plip[3] == 0))
                  planetIP = null;
                else
                  planetIP = InetAddress.getByAddress(plip);
                  
                nextPlanetReportMillis = nextSunReportMillis = now;
              }
            }
            
            if ((planetIP != null) && pack.getAddress().equals(planetIP) && (pack.getPort() == planetPort)) {
              int packlen = pack.getLength();
              int rid = decodeShort(pack.getData(), 0);
              totalBytesReceived += 28 + packlen;
              
              for (int offset=0; offset<=(packlen-4); offset+=4) {
                int id = decodeShort(pack.getData(), offset);
                int responseSize = decodeShort(pack.getData(), offset+2);

                if (id == 0) /* the list is terminated by a zero */
                  break;

                if (id == 0xFFFE) {
                  for (int i=0; i<PROBE_HISTORY_LENGTH; i++)
                    probeHistory[i] = 0;
                    
                  System.arraycopy(pack.getData(), 0, outputBuffer, 0, packlen);
                  probeIdsInHistory = 0;
                  break;
                }

                boolean seenBefore = false;
                for (int i=0; i<PROBE_HISTORY_LENGTH; i++) {
                  if (probeHistory[i] == id)
                    seenBefore = true;
                }
                
                if (!seenBefore || (id == 0xFFFF)) {
                  if (responseSize >= 2)
                    encodeShort(rid, outputBuffer, 0);

                  if (responseSize >= 4)
                    encodeShort(id, outputBuffer, 2);
                
                  int numIdsToInclude = (responseSize-4)/2;
                  if (numIdsToInclude > PROBE_HISTORY_MAX_INCLUDE)
                    numIdsToInclude = PROBE_HISTORY_MAX_INCLUDE;
                  if (numIdsToInclude > probeIdsInHistory)
                    numIdsToInclude = probeIdsInHistory;
                    
                  for (int i=0; i<numIdsToInclude; i++)
                    encodeShort(probeHistory[i], outputBuffer, 4+2*i);
                    
                  encodeShort(0, outputBuffer, 4+2*numIdsToInclude);
                  encodeInt((int)recvTS, outputBuffer, 6+2*numIdsToInclude);
                  encodeInt((int)(System.nanoTime()/1000), outputBuffer, 10+2*numIdsToInclude);
                    
                  if (responseSize > 0)  {
                    totalBytesSent += 28+responseSize;
                    ds.send(new DatagramPacket(outputBuffer, responseSize, pack.getSocketAddress()));
                  }
                  
                  if (id != 0xFFFF) {
                    for (int i=PROBE_HISTORY_LENGTH-2; i>=0; i--)
                      probeHistory[i+1] = probeHistory[i];
                    probeHistory[0] = id;
                    if (probeIdsInHistory < PROBE_HISTORY_LENGTH)
                      probeIdsInHistory ++;
                  }
                }
              }
            }
          } catch (SocketTimeoutException ste) {
            /* this is expected when the timeout expires */
          }
        } catch (IOException ioe) {
          System.out.println("IOE: "+ioe);
          ioe.printStackTrace();
        }
      }
    } catch (Exception e) {
      System.out.println("E: "+e);
      e.printStackTrace();
    }
  }
	
  public static void main(String[] args) {
    if (args.length != 4) {
      System.err.println("Syntax: Satellite <email> <identifier> <sunip> <sunport>");
      System.exit(1);
    }
    
    JFrame frame = new JFrame("SatelliteLab");
    frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    frame.pack();
    frame.setVisible(false);
    
    try {
      Satellite sat = new Satellite(args[0], args[1], InetAddress.getByName(args[2]), Integer.valueOf(args[3]).intValue());
      sat.start();
    } catch (UnknownHostException uhe) {
      System.err.println("Cannot resolve planet address: "+args[2]+":"+args[3]);
    } catch (UnsupportedEncodingException use) { }
    System.exit(0);
  }
}

